#!/usr/bin/env python3
"""Runs mDNS zeroconf service for Home Assistant discovery."""
import logging
import socket
from typing import Optional

_LOGGER = logging.getLogger(__name__)

try:
    from zeroconf.asyncio import AsyncServiceInfo, AsyncZeroconf
except ImportError:
    _LOGGER.fatal("pip install zeroconf")
    raise

MDNS_TARGET_IP = "224.0.0.251"


async def register_server(name: str, port: int, host: Optional[str] = None) -> None:
    if not host:
        test_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        test_sock.setblocking(False)
        test_sock.connect((MDNS_TARGET_IP, 1))
        host = test_sock.getsockname()[0]
        _LOGGER.debug("Detected IP: %s", host)

    assert host

    service_info = AsyncServiceInfo(
        "_wyoming._tcp.local.",
        f"{name}._wyoming._tcp.local.",
        addresses=[socket.inet_aton(host)],
        port=port,
    )
    aiozc = AsyncZeroconf()
    await aiozc.async_register_service(service_info)



"""Test that eventable classes are correct."""

import importlib
import inspect
import pkgutil
from typing import Any, Dict, List, Type

import pytest

import wyoming
from wyoming.audio import AudioFormat
from wyoming.event import Eventable
from wyoming.info import (
    AsrModel,
    AsrProgram,
    Attribution,
    HandleModel,
    HandleProgram,
    IntentModel,
    IntentProgram,
    MicProgram,
    Satellite,
    SndProgram,
    TtsProgram,
    TtsVoice,
    TtsVoiceSpeaker,
    WakeModel,
    WakeProgram,
)
from wyoming.intent import Entity
from wyoming.pipeline import PipelineStage
from wyoming.tts import SynthesizeVoice


def all_unique_eventables() -> List[Type[Eventable]]:
    found_classes = set()
    for _finder, name, _ispkg in pkgutil.walk_packages(
        wyoming.__path__, wyoming.__name__ + "."
    ):
        module = importlib.import_module(name)
        for _cls_name, cls_obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(cls_obj, Eventable) and cls_obj is not Eventable:
                found_classes.add(cls_obj)  # set for uniqueness

    return list(found_classes)


EVENTABLE_CLASSES = all_unique_eventables()

TEST_NAME = "test-name"
TEST_TEXT = "test text"
TEST_CONTEXT = {"test": "context"}
TEST_AUDIO_SETTINGS = {"rate": 22050, "width": 2, "channels": 1}
TEST_AUDIO_FORMAT = AudioFormat(22050, 2, 1)
TEST_ID = "test-id"
TEST_LANGUAGE = "test-language"
TEST_SPEAKER = "test-speaker"
TEST_TIMESTAMP = 1234
TEST_URL = "test://url"
TEST_DESCRIPTION = "test description"
TEST_VERSION = "test-version"
TEST_WAKE_WORD = "test-wake-word"
TEST_VOICE = SynthesizeVoice(name=TEST_NAME, speaker=TEST_SPEAKER)
TEST_ATTRIBUTION = Attribution(name=TEST_NAME, url=TEST_URL)

TEST_DATA: Dict[str, Dict[str, Any]] = {
    # info
    "Describe": {},
    "Info": {
        "asr": [
            AsrProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    AsrModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                    )
                ],
                supports_transcript_streaming=True,
            )
        ],
        "tts": [
            TtsProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                voices=[
                    TtsVoice(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                        speakers=[TtsVoiceSpeaker(TEST_SPEAKER)],
                    )
                ],
                supports_synthesize_streaming=True,
            )
        ],
        "handle": [
            HandleProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    HandleModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                    )
                ],
                supports_handled_streaming=True,
            )
        ],
        "intent": [
            IntentProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    IntentModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                    )
                ],
            )
        ],
        "wake": [
            WakeProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    WakeModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                        phrase="test phrase",
                    )
                ],
            )
        ],
        "mic": [
            MicProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                mic_format=TEST_AUDIO_FORMAT,
            )
        ],
        "snd": [
            SndProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                snd_format=TEST_AUDIO_FORMAT,
            )
        ],
        "satellite": Satellite(
            name=TEST_NAME,
            attribution=TEST_ATTRIBUTION,
            installed=True,
            description=TEST_DESCRIPTION,
            version=TEST_VERSION,
            area="test area",
            has_vad=True,
            active_wake_words=[TEST_WAKE_WORD],
            max_active_wake_words=1,
            supports_trigger=True,
        ),
    },
    # audio
    "AudioStart": TEST_AUDIO_SETTINGS,
    "AudioChunk": {**TEST_AUDIO_SETTINGS, "audio": bytes(100)},
    "AudioStop": {},
    # vad
    "VoiceStarted": {"timestamp": TEST_TIMESTAMP},
    "VoiceStopped": {"timestamp": TEST_TIMESTAMP},
    # wake
    "Detect": {"names": [TEST_NAME], "context": TEST_CONTEXT},
    "Detection": {
        "name": TEST_NAME,
        "timestamp": TEST_TIMESTAMP,
        "speaker": TEST_SPEAKER,
        "context": TEST_CONTEXT,
    },
    "NotDetected": {"context": TEST_CONTEXT},
    # asr
    "Transcribe": {
        "name": TEST_NAME,
        "context": TEST_CONTEXT,
        "language": TEST_LANGUAGE,
    },
    "Transcript": {
        "text": TEST_TEXT,
        "context": TEST_CONTEXT,
        "language": TEST_LANGUAGE,
    },
    "TranscriptStart": {
        "context": TEST_CONTEXT,
        "language": TEST_LANGUAGE,
    },
    "TranscriptChunk": {"text": TEST_TEXT},
    "TranscriptStop": {},
    # intent
    "Recognize": {"text": TEST_TEXT, "context": TEST_CONTEXT},
    "Intent": {
        "name": "TestIntent",
        "entities": [Entity("test entity", "test-value")],
        "context": TEST_CONTEXT,
    },
    "NotRecognized": {"text": TEST_TEXT, "context": TEST_CONTEXT},
    # handle
    "Handled": {
        "text": TEST_TEXT,
        "context": TEST_CONTEXT,
    },
    "HandledStart": {"context": TEST_CONTEXT},
    "HandledChunk": {"text": TEST_TEXT},
    "HandledStop": {},
    "NotHandled": {
        "text": TEST_TEXT,
        "context": TEST_CONTEXT,
    },
    # tts
    "SynthesizeStart": {"voice": TEST_VOICE, "context": TEST_CONTEXT},
    "SynthesizeChunk": {"text": TEST_TEXT},
    "SynthesizeStop": {},
    "SynthesizeStopped": {},
    "Synthesize": {"text": TEST_TEXT, "voice": TEST_VOICE, "context": TEST_CONTEXT},
    # timers
    "TimerStarted": {"id": TEST_ID, "total_seconds": 100},
    "TimerUpdated": {"id": TEST_ID, "total_seconds": 100, "is_active": True},
    "TimerFinished": {"id": TEST_ID},
    "TimerCancelled": {"id": TEST_ID},
    # snd
    "Played": {},
    # satellite
    "RunSatellite": {},
    "PauseSatellite": {},
    "StreamingStarted": {},
    "StreamingStopped": {},
    "SatelliteConnected": {},
    "SatelliteDisconnected": {},
    # misc
    "Error": {"text": TEST_TEXT},
    "Ping": {},
    "Pong": {},
    "RunPipeline": {"start_stage": PipelineStage.ASR, "end_stage": PipelineStage.TTS},
}


@pytest.mark.parametrize("cls", EVENTABLE_CLASSES)
def test_eventable_round_trip(cls: Type[Eventable]) -> None:
    init_kwargs = TEST_DATA[cls.__name__]
    instance = cls(**init_kwargs)

    # Test event() method
    event = instance.event()
    assert event.type is not None, f"{cls} returned event with no type"

    # Test is_type matches event.type
    assert cls.is_type(event.type), f"{cls}.is_type failed for {event.type}"

    # Test from_event returns an equivalent object
    round_trip = cls.from_event(event)
    assert round_trip == instance, f"{cls}.from_event failed to round-trip {instance}"



"""HTTP server for intent recognition/handling."""

import logging
from pathlib import Path
from typing import Any, Dict

from flask import request

from wyoming.asr import Transcript
from wyoming.client import AsyncClient
from wyoming.error import Error
from wyoming.handle import Handled, NotHandled
from wyoming.intent import Intent, NotRecognized

from .shared import get_app, get_argument_parser

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "intent.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--language", help="Language for text")
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("intent", CONF_PATH, args)

    @app.route("/api/recognize-intent", methods=["POST", "GET"])
    async def api_stt() -> Dict[str, Any]:
        uri = request.args.get("uri", args.uri)
        if not uri:
            raise ValueError("URI is required")

        if request.method == "POST":
            text = request.data.decode()
        else:
            text = request.args.get("text", "")

        if not text:
            raise ValueError("Text is required")

        language = request.args.get("language", args.language)

        async with AsyncClient.from_uri(uri) as client:
            await client.write_event(Transcript(text=text, language=language).event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                success = False
                type_name = "unknown"
                result: Dict[str, Any] = {}

                if Intent.is_type(event.type):
                    success = True
                    type_name = "intent"
                    intent = Intent.from_event(event)
                    result = intent.to_dict()
                elif Handled.is_type(event.type):
                    success = True
                    type_name = "handled"
                    handled = Handled.from_event(event)
                    result = handled.to_dict()
                elif NotRecognized.is_type(event.type):
                    success = False
                    type_name = "not-recognized"
                    not_recognized = NotRecognized.from_event(event)
                    result = not_recognized.to_dict()
                elif NotHandled.is_type(event.type):
                    success = False
                    type_name = "not-handled"
                    not_handled = NotHandled.from_event(event)
                    result = not_handled.to_dict()
                elif Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

                return {"success": success, "type": type_name, "result": result}

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()



"""Information about available services, models, etc.."""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from .audio import AudioFormat
from .event import Event, Eventable
from .util.dataclasses_json import DataClassJsonMixin

DOMAIN = "info"
_DESCRIBE_TYPE = "describe"
_INFO_TYPE = "info"


@dataclass
class Describe(Eventable):
    """Request info message."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _DESCRIBE_TYPE

    def event(self) -> Event:
        return Event(type=_DESCRIBE_TYPE)

    @staticmethod
    def from_event(event: Event) -> "Describe":
        return Describe()


@dataclass
class Attribution(DataClassJsonMixin):
    """Attribution for an artifact."""

    name: str
    """Who made it."""

    url: str
    """Where it's from."""


@dataclass
class Artifact(DataClassJsonMixin):
    """Information about a service, model, etc.."""

    name: str
    """Name/id of artifact."""

    attribution: Attribution
    """Who made the artifact and where it's from."""

    installed: bool
    """True if the artifact is currently installed."""

    description: Optional[str]
    """Human-readable description of the artifact."""

    version: Optional[str]
    """Version of the artifact."""


# -----------------------------------------------------------------------------


@dataclass
class AsrModel(Artifact):
    """Speech-to-text model."""

    languages: List[str]
    """List of supported model languages."""


@dataclass
class AsrProgram(Artifact):
    """Speech-to-text service."""

    models: List[AsrModel]
    """List of available models."""

    supports_transcript_streaming: bool = False
    """True if transcript streaming events are supported."""


# -----------------------------------------------------------------------------


@dataclass
class TtsVoiceSpeaker(DataClassJsonMixin):
    """Individual speaker in a multi-speaker voice."""

    name: str
    """Name/id of speaker."""


@dataclass
class TtsVoice(Artifact):
    """Text-to-speech voice."""

    languages: List[str]
    """List of languages available in the voice."""

    speakers: Optional[List[TtsVoiceSpeaker]] = None
    """List of individual speakers in the voice."""


@dataclass
class TtsProgram(Artifact):
    """Text-to-speech service."""

    voices: List[TtsVoice]
    """List of available voices."""

    supports_synthesize_streaming: bool = False
    """True if synthesize text streaming events are supported."""


# -----------------------------------------------------------------------------


@dataclass
class HandleModel(Artifact):
    """Intent handling model."""

    languages: List[str]
    """List of supported languages in the model."""


@dataclass
class HandleProgram(Artifact):
    """Intent handling service."""

    models: List[HandleModel]
    """List of available models."""

    supports_handled_streaming: bool = False
    """True if handled response streaming events are supported."""


# -----------------------------------------------------------------------------


@dataclass
class WakeModel(Artifact):
    """Wake word detection model."""

    languages: List[str]
    """List of languages supported by the model."""

    phrase: Optional[str]
    """Wake up phrase used by the model."""


@dataclass
class WakeProgram(Artifact):
    """Wake word detection service."""

    models: List[WakeModel]
    """List of available models."""


# -----------------------------------------------------------------------------


@dataclass
class IntentModel(Artifact):
    """Intent recognition model."""

    languages: List[str]
    """List of languages supported by the model."""


@dataclass
class IntentProgram(Artifact):
    """Intent recognition service."""

    models: List[IntentModel]
    """List of available models."""


# -----------------------------------------------------------------------------


@dataclass
class Satellite(Artifact):
    """Satellite information."""

    area: Optional[str] = None
    """Name of the area the satellite is in."""

    has_vad: Optional[bool] = None
    """True if a local VAD will be used to detect the end of voice commands."""

    active_wake_words: Optional[List[str]] = None
    """Wake words that are currently being listened for."""

    max_active_wake_words: Optional[int] = None
    """Maximum number of local wake words that can be run simultaneously."""

    supports_trigger: Optional[bool] = None
    """Satellite supports remotely triggering pipeline runs."""


# -----------------------------------------------------------------------------


@dataclass
class MicProgram(Artifact):
    """Microphone information."""

    mic_format: AudioFormat
    """Input audio format."""


# -----------------------------------------------------------------------------


@dataclass
class SndProgram(Artifact):
    """Sound output information."""

    snd_format: AudioFormat
    """Output audio format."""


# -----------------------------------------------------------------------------


@dataclass
class Info(Eventable):
    """Response to describe message with information about available services, models, etc."""

    asr: List[AsrProgram] = field(default_factory=list)
    """Speech-to-text services."""

    tts: List[TtsProgram] = field(default_factory=list)
    """Text-to-speech services."""

    handle: List[HandleProgram] = field(default_factory=list)
    """Intent handling services."""

    intent: List[IntentProgram] = field(default_factory=list)
    """Intent recognition services."""

    wake: List[WakeProgram] = field(default_factory=list)
    """Wake word detection services."""

    mic: List[MicProgram] = field(default_factory=list)
    """Audio input services."""

    snd: List[SndProgram] = field(default_factory=list)
    """Audio output services."""

    satellite: Optional[Satellite] = None
    """Satellite information."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _INFO_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {
            "asr": [p.to_dict() for p in self.asr],
            "tts": [p.to_dict() for p in self.tts],
            "handle": [p.to_dict() for p in self.handle],
            "intent": [p.to_dict() for p in self.intent],
            "wake": [p.to_dict() for p in self.wake],
            "mic": [p.to_dict() for p in self.mic],
            "snd": [p.to_dict() for p in self.snd],
        }

        if self.satellite is not None:
            data["satellite"] = self.satellite.to_dict()

        return Event(type=_INFO_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Info":
        assert event.data is not None

        satellite: Optional[Satellite] = None
        satellite_data = event.data.get("satellite")
        if satellite_data is not None:
            satellite = Satellite.from_dict(satellite_data)

        return Info(
            asr=[AsrProgram.from_dict(d) for d in event.data.get("asr", [])],
            tts=[TtsProgram.from_dict(d) for d in event.data.get("tts", [])],
            handle=[HandleProgram.from_dict(d) for d in event.data.get("handle", [])],
            intent=[IntentProgram.from_dict(d) for d in event.data.get("intent", [])],
            wake=[WakeProgram.from_dict(d) for d in event.data.get("wake", [])],
            mic=[MicProgram.from_dict(d) for d in event.data.get("mic", [])],
            snd=[SndProgram.from_dict(d) for d in event.data.get("snd", [])],
            satellite=satellite,
        )



#!/usr/bin/env python3

import argparse
import asyncio
import logging
import time
from dataclasses import dataclass
from functools import partial
from typing import List, Optional, Set

from .asr import Transcript
from .client import AsyncClient
from .event import Event
from .info import AsrModel, AsrProgram, Attribution, Describe, Info
from .server import AsyncEventHandler, AsyncServer

_LOGGER = logging.getLogger(__name__)


async def main() -> None:
    """Runs fallback ASR server."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--uri", required=True, help="unix:// or tcp://")
    parser.add_argument(
        "fallback_uri",
        nargs="+",
        help="Wyoming URI of ASR server (e.g., tcp://127.0.0.1:10300)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to console"
    )
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    _LOGGER.debug(args)

    server = AsyncServer.from_uri(args.uri)
    _LOGGER.info("Ready")

    try:
        await server.run(partial(FallbackEventHandler, args.fallback_uri))
    except KeyboardInterrupt:
        pass


# -----------------------------------------------------------------------------


@dataclass
class FallbackClient:
    """Information about a Wyoming ASR client."""

    uri: str
    client: AsyncClient
    name: str
    info: Optional[Info] = None
    is_connected: bool = False
    is_enabled: bool = True

    async def ensure_connected_and_info(self) -> None:
        """Connect to client and get info if necessary."""
        if not self.is_connected:
            _LOGGER.debug("Connecting to %s", self.uri)
            await self.client.connect()
            self.is_connected = True

        if self.info is None:
            _LOGGER.debug("Getting info for %s", self.uri)
            await self.client.write_event(Describe().event())
            while True:
                event = await self.client.read_event()
                if event is None:
                    _LOGGER.warning("Disconnected from %s", self.name)
                    self.is_enabled = False
                    self.is_connected = False
                    break

                if not Info.is_type(event.type):
                    continue

                self.info = Info.from_event(event)
                if self.info.asr:
                    self.name = self.info.asr[0].name
                    self.is_enabled = True
                else:
                    _LOGGER.warning("Disabled %s (no ASR service)", self.name)
                    self.is_enabled = False

                break

            _LOGGER.info("Got info for %s (%s)", self.name, self.uri)


class FallbackEventHandler(AsyncEventHandler):
    """Event handler for clients."""

    def __init__(
        self,
        fallback_uris: List[str],
        *args,
        **kwargs,
    ) -> None:
        """Initialize event handler."""
        super().__init__(*args, **kwargs)

        if not fallback_uris:
            raise ValueError("At least one fallback URI is required")

        self.fallback_clients: List[FallbackClient] = [
            FallbackClient(uri=uri, client=AsyncClient.from_uri(uri), name=uri)
            for uri in fallback_uris
        ]
        self.client_id = str(time.monotonic_ns())

        self._target_idx = 0
        self._target_read_task: Optional[asyncio.Task] = None
        self._saved_events: List[Event] = []
        self._info_event: Optional[Event] = None

    async def handle_event(self, event: Event) -> bool:
        """Handle Wyoming event."""
        try:
            if Describe.is_type(event.type):
                await self._write_info()
                return True

            target_client = self.fallback_clients[self._target_idx]
            await target_client.ensure_connected_and_info()

            if self._target_read_task is None:
                await self._set_fallback_target(target_client)

            # Forward to current target
            await target_client.client.write_event(event)
            self._saved_events.append(event)
        except Exception:
            _LOGGER.exception("Error handling event")

        return True

    async def _set_fallback_target(self, fb_client: FallbackClient) -> None:
        """Create read task and send saved messages."""
        self._target_read_task = asyncio.create_task(self._read_from_client(fb_client))
        self._target_read_task.add_done_callback(lambda _: self._clear_read_task())

        for saved_event in self._saved_events:
            await fb_client.client.write_event(saved_event)

    async def _read_from_client(self, fb_client: FallbackClient) -> None:
        """Read events from a client until a transcript comes back."""
        try:
            while True:
                event = await fb_client.client.read_event()
                if event is None:
                    # TODO: fall back
                    _LOGGER.warning("Disconnected from %s", fb_client.name)
                    # self.is_connected = False
                    break

                if not Transcript.is_type(event.type):
                    continue

                transcript = Transcript.from_event(event)
                _LOGGER.debug("Received from '%s' from %s", transcript, fb_client.name)

                await fb_client.client.disconnect()
                fb_client.is_connected = False

                if transcript.text.strip():
                    # Got the final transcript
                    _LOGGER.debug(
                        "Final transcript: '%s' from %s",
                        transcript.text,
                        fb_client.name,
                    )
                    await self.write_event(event)
                    self._saved_events.clear()
                    self._target_idx = 0
                else:
                    # Fallback
                    self._target_idx += 1
                    while self._target_idx < len(self.fallback_clients):
                        next_client = self.fallback_clients[self._target_idx]
                        await next_client.ensure_connected_and_info()
                        if not next_client.is_enabled:
                            self._target_idx += 1
                            continue

                        _LOGGER.debug("Falling back to %s", next_client.name)
                        await self._set_fallback_target(next_client)
                        break

                    if self._target_idx >= len(self.fallback_clients):
                        # No more fallbacks
                        _LOGGER.debug("No more fallbacks. Sending empty transcript.")
                        await self.write_event(event)
                        self._saved_events.clear()
                        self._target_idx = 0

                break

        except Exception:
            _LOGGER.exception("Error reading event from client")

    def _clear_read_task(self) -> None:
        self._target_read_task = None

    async def _write_info(self) -> None:
        if self._info_event is not None:
            await self.write_event(self._info_event)
            return

        supported_langs: Set[str] = set()

        for fb_client in self.fallback_clients:
            await fb_client.ensure_connected_and_info()
            assert fb_client.info is not None
            for asr_prog in fb_client.info.asr:
                for asr_model in asr_prog.models:
                    if not asr_model.installed:
                        continue

                    supported_langs.update(asr_model.languages)

        info = Info(
            asr=[
                AsrProgram(
                    name="stt-fallback",
                    attribution=Attribution(
                        name="The Home Assistant Authors",
                        url="http://github.com/OHF-voice",
                    ),
                    description="Automatic fallback for Wyoming speech-to-text",
                    installed=True,
                    version="1.0.0",
                    models=[
                        AsrModel(
                            name="fallback",
                            attribution=Attribution(
                                name="The Home Assistant Authors",
                                url="http://github.com/OHF-voice",
                            ),
                            installed=True,
                            description="Fallback model",
                            version=None,
                            languages=list(sorted(supported_langs)),
                        )
                    ],
                )
            ]
        )

        self._info_event = info.event()
        await self.write_event(self._info_event)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    asyncio.run(main())
